// cwasm.cpp : definition of the Redcode Assembler
/////////////////////////////////////////////////////////////////////////////


#include "afx.h"		// File/Archive Definitions

#include "cwasm.h"

#include "ctype.h"		// isalpha, isdigit functions prototypes
#include "cwdlg.h"		// Dialog Window Definitions


IMPLEMENT_DYNCREATE(CRedcodeAssembler, CObject)


/////////////////////////////////////////////////////////////////////////////
// Construction & Destruction
CRedcodeAssembler::CRedcodeAssembler()
{
	// Clear All Data
	m_EQU_Table.RemoveAll();
	m_RIF_Table.RemoveAll();
	m_CROSS_Table.RemoveAll();
}

CRedcodeAssembler::~CRedcodeAssembler()
{
}


/////////////////////////////////////////////////////////////////////////////
// Token Extraction
BOOL CRedcodeAssembler::CharInSet(const char ch, const char* pSet)
{
	int i = 0;
	int chars = _fstrlen(pSet);
	
	while ((i<chars) && (ch!=pSet[i]))
		i++;
	
	return (i<chars);
}

CString CRedcodeAssembler::ExtractNextToken(CString* pLine,
											const char* pFrontSeps,
											const char* pRearSeps)
{
	// Different Separators for
	// Token Head-Tile Search ?
	if (pRearSeps==NULL)	
		pRearSeps = pFrontSeps;
	
	CString token;      
	int nChars = pLine->GetLength();
	
	// Jump Over Left Separators
	int nCount = 0;
	while ((nCount<nChars)
			&& (CharInSet(pLine->GetAt(nCount),pFrontSeps)))
		nCount++;
		
	int tkStart = nCount;
	
	// Walk Until Another Separator is Found
	while ((nCount<nChars)
			&& (!CharInSet(pLine->GetAt(nCount),pRearSeps)))
		nCount++;
		
	int tkEnd = nCount;
	
	// Remove the Token Form the Source String
	token = pLine->Left(tkEnd);
	token = token.Mid(tkStart);
	*pLine = pLine->Mid(tkEnd);

	return token;
}

BOOL CRedcodeAssembler::IsTokenValidLabel(const char* pToken)
{
	int len = _fstrlen(pToken);
	
	if (len<=0)
		return FALSE;
		
	// First Char Must Be a Letter
	if (!isalpha(pToken[0]))
		return FALSE;
		
	int i = 1;				
	while ((i<len) && (isalnum(pToken[i])))
		i++;

	// TRUE only if scan is completed
	return (i>=len);			
}

BOOL CRedcodeAssembler::IsTokenSignedInteger(const char* pToken)
{
	int i = 0;
	int len = _fstrlen(pToken);
	
	// Step Over Sign
	if ((len>0)
		&& ((pToken[0]=='+') || (pToken[0]=='-')))
		i++;
	
	while ((i<len) && (isdigit(pToken[i])))
		i++;

	// TRUE only if all chars are '0'..'9'
	return (i>=len);			
}


/////////////////////////////////////////////////////////////////////////////
// Line Assembling
int CRedcodeAssembler::IsRedcodeStatement(const char* pToken)
{
	// Search in the "HandBook"	
	extern const char* COMMANDS[];
	extern const int COMMANDS_ITEMS;
	
	int count = 0;
	
	while ((count<COMMANDS_ITEMS)
			&& (_fstricmp(pToken,COMMANDS[count])!=0))
		count++;
		
	if (count<COMMANDS_ITEMS)		
		return count;			// Return Statement Code
	else
		return -1;				// Instruction Not Found
}

BOOL CRedcodeAssembler::AssembleOperand(const char* pToken,
										const char* pOperand,
										OP_MODE* pOpMode,
										OP_DATA* pOpData,
										char* pErrorString)
{
	CString token = pToken;

	// Empty Token?
	if (token.GetLength()==0)
	{
		// Operand Expected
		if (pErrorString!=NULL)
			_fstrcpy(pErrorString,ERR_MSG_OPERAND_EXPECTED);
			
		return FALSE;
	}
	
	// Default Addressing Mode
	extern const char ADDR_MODES[];
	extern const int ADDR_MODES_ITEMS;
	
	if (!CharInSet(token[0],ADDR_MODES))
		token = ADDR_REL_DIRECT + token;
		
	*pOpMode = 0;
	
	while ((*pOpMode<ADDR_MODES_ITEMS)
			&& (token[0]!=ADDR_MODES[*pOpMode]))
		(*pOpMode)++;                          
		
	if (*pOpMode>=ADDR_MODES_ITEMS)
	{
		// Invalid Addressing Mode 
		if (pErrorString!=NULL)
			_fstrcpy(pErrorString,ERR_MSG_INVALID_ADDR_MODE);
			
		return FALSE;
	}

	token = token.Mid(1);

	if (token.GetLength()==0)
	{
		// Operand Expected
		if (pErrorString!=NULL)
			_fstrcpy(pErrorString,ERR_MSG_OPERAND_EXPECTED);
			
		return FALSE;
	}
	
	if (IsTokenSignedInteger(token))
	{
		*pOpData = (OP_DATA)atoi(token);
	}   
	else
	{
		// Maybe A Label...
		if (!IsTokenValidLabel(token))
		{
			// Invalid Label
			if (pErrorString!=NULL)
				_fstrcpy(pErrorString,ERR_MSG_INVALID_LABEL);
		
			return FALSE;
		}

		// Table This Cross-Rif
		char buffer[16];
		CString rif = _itoa(m_nObjCodeLine,buffer,10);
		rif += pOperand;
		
		m_CROSS_Table[rif] = token;
	}

	return TRUE;
}

BOOL CRedcodeAssembler::ValidateAddrModes(CCoreCell* pNewCell,
										  char* pErrorString)
{
	switch (pNewCell->GetCommand())
	{
		case 0 :	// DAT B
					if ((pNewCell->GetOpBMode()==2)
						|| (pNewCell->GetOpBMode()==3))
					{
						// DAT Wants Only Immediate or Direct A.M.
						if (pErrorString!=NULL)
							_fstrcpy(pErrorString,ERR_MSG_INVALID_ADDR_MODE);
						
						return FALSE;
					}
					break;
		
		case 1 :	// MOV A,B        
		case 2 :	// ADD A,B	
		case 3 :	// SUB A,B
					if (pNewCell->GetOpBMode()==0)
					{
						// B Must Not Be Immediate
						if (pErrorString!=NULL)
							_fstrcpy(pErrorString,ERR_MSG_INVALID_ADDR_MODE);
						
						return FALSE;
					}
					break;

		case 4 :	// JMP A
					if (pNewCell->GetOpAMode()==0)
					{
						// A Must Not Be Immediate
						if (pErrorString!=NULL)
							_fstrcpy(pErrorString,ERR_MSG_INVALID_ADDR_MODE);
						
						return FALSE;
					}
					break;

		case 5 :	// JMZ A,B	
		case 6 :	// JMN A,B
		case 7 :	// DJN A,B
					if (pNewCell->GetOpAMode()==0)
					{
						// A Must Not Be Immediate
						if (pErrorString!=NULL)
							_fstrcpy(pErrorString,ERR_MSG_INVALID_ADDR_MODE);
						
						return FALSE;
					}
					break;

		case 8 :	// CMP A,B
					// All Forms Are Allowed...
					break;
		
		case 9 :	// ---
					break;
		
		case 10 :	// SPL B
					if (pNewCell->GetOpBMode()==0)
					{
						// B Must Not Be Immediate
						if (pErrorString!=NULL)
							_fstrcpy(pErrorString,ERR_MSG_INVALID_ADDR_MODE);
						
						return FALSE;
					}
					break;
		default :	
					break;
	}

	return TRUE;
}												

BOOL CRedcodeAssembler::AssembleLine(const char* pLine, char* pErrorString)
{
	// Copy Source Text
	CString line = pLine;
	CCoreCell* pNewCell = NULL;
	
	// Get First Token
	extern const char FIELD_SEPS[];
	extern const char OPERAND_SEPS[];
	extern const char BEGIN_COMMENT[];

	CString token = ExtractNextToken(&line,FIELD_SEPS);
	
	// Empty Line or Comment?
	if ((token.GetLength()==0)
		|| (CharInSet(token[0],BEGIN_COMMENT)))
		return TRUE;

	// Statement?
	if (token.CompareNoCase(EQU_STATEMENT)==0)
	{
		// Label Expected 
		if (pErrorString!=NULL)
			_fstrcpy(pErrorString,ERR_MSG_LABEL_EXPECTED);
		
		return FALSE;
	}

	// Search for a Label
	CString label;
	
	if (IsRedcodeStatement(token)==-1)
	{
		// Maybe a Label...
		if (!IsTokenValidLabel(token))			
		{
			// Invalid Label (No Alphabetical Chars Used)
			if (pErrorString!=NULL)
				_fstrcpy(pErrorString,ERR_MSG_INVALID_LABEL);
			
			return FALSE;
		}
	
		// Label Found!
		CString aux;
		label = token;                  

		if ((m_EQU_Table.Lookup(label,aux))
			|| (m_RIF_Table.Lookup(label,aux)))
		{
			// Duplicated Label
			if (pErrorString!=NULL)
				_fstrcpy(pErrorString,ERR_MSG_DUPLICATED_LABEL);
			
			return FALSE;			
		}
		
		// Read Redcode Instruction
		token = ExtractNextToken(&line,FIELD_SEPS);
		
		if (token.CompareNoCase(EQU_STATEMENT)==0)
		{
			if (label.CompareNoCase(START_LABEL)==0)
			{
				// Invalid "Start" Label Placement
				if (pErrorString!=NULL)
					_fstrcpy(pErrorString,ERR_MSG_INVALID_START_POS);
			
				return FALSE;			
			}
		}
		else
			if (IsRedcodeStatement(token)==-1)
			{
				// Redcode Statement Expected
				if (pErrorString!=NULL)
					_fstrcpy(pErrorString,ERR_MSG_STATEMENT_EXPECTED);
		
				return FALSE;
			}
			else
			{
				// Surely <token> is a Redcode Instruction
				// Insert this Label in the Rif. Table
				char buffer[16];
				m_RIF_Table[label] = _itoa(m_nObjCodeLine,buffer,10);
		
				if (label.CompareNoCase(START_LABEL)==0)
				{
					// Found Starting Point
					m_nStartLine = m_nObjCodeLine;
				}
			}
	}

	// token = Redcode Symbolic Statement (included EQU)
	if (token.CompareNoCase(EQU_STATEMENT)==0)	
	{
		// Read Numeric Value
		token = ExtractNextToken(&line,FIELD_SEPS);
		
		if (!IsTokenSignedInteger(token))
		{
			// Number Expected
			if (pErrorString!=NULL)
				_fstrcpy(pErrorString,ERR_MSG_NUMBER_EXPECTED);
			
			return FALSE;
		}
		
		// Table Label & Related Value
		m_EQU_Table[label] = token;
	}
	else
	{
		// Redcode Statement Found
		pNewCell = new CCoreCell;		
		
		if (pNewCell==0)
		{
			if (pErrorString!=NULL)
				_fstrcpy(pErrorString,ERR_MSG_MEMORY_FULL);

			return FALSE;
		}                                                
		
		// Assemble Operative Code
		pNewCell->SetCommand((OP_CODE)IsRedcodeStatement(token));
	
		// Read First Operand
		OP_MODE opMode;
		OP_DATA opData;
		CString operand;
		token = ExtractNextToken(&line,FIELD_SEPS,OPERAND_SEPS);
		
		if ((pNewCell->GetCommand()==0)			// DAT
			|| (pNewCell->GetCommand()==10))	// SPL
		{
			// For DAT & SPL, B Is the First Operand
			operand = B_OPERAND_CHR_ID;
		}
		else
		{
			// For Other Statements, the First Operand Is A
			operand = A_OPERAND_CHR_ID;
		}
		
		if (!AssembleOperand(token,operand,&opMode,&opData,pErrorString))
		{
			// Invalid Operand
			return FALSE;
		}
		else
		{
			if (operand[0]==A_OPERAND_CHR_ID)
			{
				pNewCell->SetOpAMode(opMode);
				pNewCell->SetOpAValue(opData);
			}
			else
			{
				pNewCell->SetOpBMode(opMode);
				pNewCell->SetOpBValue(opData);
        	}
		}
		
		// Search for the Second Operand
		if ((pNewCell->GetCommand()!=0)			// DAT
			&& (pNewCell->GetCommand()!=4)		// JMP
			&& (pNewCell->GetCommand()!=10))	// SPL   
		{
			token = ExtractNextToken(&line,OPERAND_SEPS,FIELD_SEPS);
			operand = B_OPERAND_CHR_ID;

			if (!AssembleOperand(token,operand,&opMode,&opData,pErrorString))
			{
				// Invalid Operand
				return FALSE;
			}

			pNewCell->SetOpBMode(opMode);
			pNewCell->SetOpBValue(opData);
		}

		// Validate Addressing Modes
		if (!ValidateAddrModes(pNewCell,pErrorString))
		{
			// Invalid Addressing Mode
			return FALSE;
		}
	}
	
	// Read Last token
	token = ExtractNextToken(&line,FIELD_SEPS);
	
	if ((token.GetLength()>0)
		&& (!CharInSet(token[0],BEGIN_COMMENT)))
	{
		// A Comment Was Expected
		if (pErrorString!=NULL)
			_fstrcpy(pErrorString,ERR_MSG_COMMENT_EXPECTED);
		
		return FALSE;
	}

	// The Line Has Been Assembled Correctly!
	// Add the Assembled Instruction to the Encoded Program
	if (pNewCell!=NULL)
	{
		m_nObjCodeLine++;
		m_object.AddStatement(pNewCell);

		if (m_nObjCodeLine>=MAX_CORE_PROGRAMS_LENGTH)
		{
			// Program Too Long
			if (pErrorString!=NULL)
				_fstrcpy(pErrorString,ERR_MSG_PROGRAM_TOO_LONG);
			
			return FALSE;
		}
	}
	
	return TRUE;
}


/////////////////////////////////////////////////////////////////////////////
// Redcode Assembler
BOOL CRedcodeAssembler::FirstAssemblingStep(char* pErrorString, int* pLine)
{
	// Start From the Beginning...
	int nLine = 0;
	CString textLine;
	BOOL done = TRUE;
	POSITION pos = m_source->GetHeadPosition();
	
	while ((pos!=NULL) && (done))
	{
		textLine = m_source->GetNext(pos);
		done = AssembleLine(textLine,pErrorString);
		nLine++;
	}

	// Return Wrong Line Index
	if ((!done) && (pLine!=NULL))
		*pLine = nLine;
		
	return done;
}

BOOL CRedcodeAssembler::SecondAssemblingStep(char* pErrorString,
											 char* pMissingLabel)
{
	int value;
	CString label;
	CString symbol;
	CString operand;
	UINT lineNumber;
	
	// Encode Symbolic Rif.     
	BOOL found = TRUE;
	POSITION pos = m_CROSS_Table.GetStartPosition();

	while ((pos!=NULL) && (found))
	{
		m_CROSS_Table.GetNextAssoc(pos,operand,label);

		// Is <label> an EQU-defined symbol?
		found = m_EQU_Table.Lookup(label,symbol);
		
		if (found)
		{
			// Retrieve Symbol Associated Value
			value = atoi(symbol);
			
			// Extract Line Number & Operand
			lineNumber = (UINT)atoi(operand.Left(operand.GetLength()-1));
			operand = operand.Right(1);
		
			// Assemble	EQU Data
			CCoreCell* pCell = m_object.GetStatement(lineNumber);
			
			if (pCell!=NULL)
			{
				if (operand[0]==A_OPERAND_CHR_ID)
					pCell->SetOpAValue(value);

				if (operand[0]==B_OPERAND_CHR_ID)
					pCell->SetOpBValue(value);
			}
		}
		else
		{
			// Is <label> an instruction pointer?
			found = m_RIF_Table.Lookup(label,symbol);
			
			if (found)
			{
				// Extract Line Number & Operand
				lineNumber = (UINT)atoi(operand.Left(operand.GetLength()-1));
				operand = operand.Right(1);

				// Calculate Line Displacement
				int value = (UINT)(atoi(symbol) - lineNumber);

				// Assemble	Line Displacement
				CCoreCell* pCell = m_object.GetStatement(lineNumber);
			
				if (pCell!=NULL)
				{
					if (operand[0]==A_OPERAND_CHR_ID)
						pCell->SetOpAValue(value);
	
					if (operand[0]==B_OPERAND_CHR_ID)
						pCell->SetOpBValue(value);
				}
			}
		}
	}
	
	// Return Error Message
	if ((!found) && (pErrorString!=NULL))
		_fstrcpy(pErrorString,ERR_MSG_LABEL_NOT_FOUND);
		
	// Return Lost Label
	if ((!found) && (pMissingLabel!=NULL))
		_fstrcpy(pMissingLabel,label);
		
	return found;
}


/////////////////////////////////////////////////////////////////////////////
// Public Interface

BOOL CRedcodeAssembler::Assemble(char* pErrorString, int* pWrongLineIndex)
{
	// Clear All Tables
	m_EQU_Table.RemoveAll();
	m_RIF_Table.RemoveAll();
	m_CROSS_Table.RemoveAll();
	
	// Reset Line Counter
	m_nStartLine = 0;
	m_nObjCodeLine = 0;
	
	// Clear Core Program
	m_object.ClearProgram();
	m_object.SetStartingPoint((CORE_ADDRESS)0);
	
	// First Pass...
	int wrongLineIndex = 0;
	if (!FirstAssemblingStep(pErrorString,&wrongLineIndex))
	{
		if (pWrongLineIndex!=NULL)
			*pWrongLineIndex = wrongLineIndex;
		return FALSE;
	}
			
	// Second Pass...
	char missingLabel[256];
	if (!SecondAssemblingStep(pErrorString,missingLabel))
	{
		if (pWrongLineIndex!=NULL)
		{
			// Search for missingLabel
			CString line;
			int lineIndex = 0;
			BOOL found = FALSE;
			POSITION pos = m_source->GetHeadPosition();

			while ((pos!=NULL) && (!found))
			{
				line = m_source->GetNext(pos);
				found = (line.Find(missingLabel)!=-1);
				lineIndex++;
			}                                         
			
			*pWrongLineIndex = lineIndex;
		}

		return FALSE;
	}
			
	if (m_nObjCodeLine==0)
	{
		if (pErrorString!=NULL)
			_fstrcpy(pErrorString,ERR_MSG_NULL_PROGRAM);
			
		// Don't Care...
		if (pWrongLineIndex!=NULL)
			*pWrongLineIndex = -1;		
	}

	// Set Starting Point
	m_object.SetStartingPoint(m_nStartLine);

	return TRUE;
}
 
void CRedcodeAssembler::LoadSourceCode(CStringList* pSourceText) 
{
	m_source = pSourceText;
} 

BOOL CRedcodeAssembler::WriteObjectCodeToDisk(const char* pSourcePathName)
{
	// Get Document File Name
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];
	char fileName[_MAX_FNAME];
	char ext[_MAX_EXT];
	
	_splitpath(pSourcePathName,drive,dir,fileName,ext);
	_fstrcat(fileName,CCodeFileDialog::defExt);
	
	// Ask for File Name
	CCodeFileDialog saveDlg("Save Coded Program",		// Window Title
							FALSE,						// Save As Operation
							CCodeFileDialog::defExt,	// Default Extension
							fileName);					// File Name
	
	if (saveDlg.DoModal()==IDOK)
	{
		// Open/Create Selected File
		CFile file;

		if (!file.Open(saveDlg.GetPathName(), CFile::modeCreate|
											  CFile::modeReadWrite|
											  CFile::shareExclusive))
			return FALSE;

		// Create Archive
		CArchive archive(&file, CArchive::store|CArchive::bNoFlushOnDelete);

		// Save Encoded Program
		HCURSOR hPrevCur = ::SetCursor(LoadCursor(NULL,IDC_WAIT));
		m_object.Serialize(archive);
	
		// Close All
		archive.Close();
		file.Close();

		::SetCursor(hPrevCur);
	}
	else
		return FALSE;

	return TRUE;
}


