// cwgraphs.h: CGraphCoreWnd definition 
/////////////////////////////////////////////////////////////////////////////


#include "resource.h"		// Window Icon
#include "cwgraphs.h" 


// Single Cell Size
#define CELL_WIDTH		3
#define CELL_HEIGHT		3

// Core Cell Status
#define NOT_USED		(BYTE)0
#define ACCESSED_P1		(BYTE)1
#define EXECUTED_P1		(BYTE)2
#define ACCESSED_P2		(BYTE)3
#define EXECUTED_P2		(BYTE)4

// Cell Colors
#define BACKGROUND_COLOR		RGB(0xff,0xff,0xff)
#define ACCESSED_P1_COLOR		RGB(0x80,0x00,0x00)
#define EXECUTED_P1_COLOR		RGB(0xff,0x00,0x00)
#define ACCESSED_P2_COLOR		RGB(0x00,0x00,0x80)
#define EXECUTED_P2_COLOR		RGB(0x00,0x00,0xff)



/////////////////////////////////////////////////////////////////////////////                                
// Advanced Mathematics
int PowOfTwo(UINT e)
{
	int aux = 1;
	while (e>0)
	{
		aux *= 2;
		e--;
	}       
	
	return aux;
}


/////////////////////////////////////////////////////////////////////////////
// CGraphCoreWnd
IMPLEMENT_DYNCREATE(CGraphCoreWnd,CMDIChildWnd) 

                      
// Default Constructor
CGraphCoreWnd::CGraphCoreWnd() : CMDIChildWnd()
{
	// Clear Core
	for (int i=0; i<MAX_CORE_SIZE; i++)
		m_coreStatus[i] = NOT_USED;
};
  


CGraphCoreWnd::~CGraphCoreWnd()
{
};
  
  
/////////////////////////////////////////////////////////////////////////////                       
// Window Creating & Destroying
BOOL CGraphCoreWnd::Create(const char*   wndTitle,
						   int           nCoreSize,
						   CMDIFrameWnd* pParentFrame,	// = NULL
						   DWORD         wndStyle)		// = 0
{                      
	// Set Correct Style
	if (wndStyle == 0)
		wndStyle = WS_CHILD | WS_VISIBLE | WS_OVERLAPPEDWINDOW;
	else
		wndStyle |= WS_CHILD;	// WS_CHILD Style Required!
		
	// No Maximized Button
	wndStyle &= ~WS_MAXIMIZEBOX;

	// Calculate Window Size
	m_coreSize = nCoreSize;
	
	int e = 0;                   
	while (PowOfTwo(e)<m_coreSize)
		e++;
	
	m_vertCells = PowOfTwo(e/2);
	m_horzCells = PowOfTwo(e - e/2);	// m_horCells >= m_vertCells
    
	CRect wndRect;   
	wndRect.top = 0;
	wndRect.left = 0;
	wndRect.right = m_horzCells*CELL_WIDTH + 1;		
	wndRect.bottom = m_vertCells*CELL_HEIGHT + 1;
	::AdjustWindowRect(wndRect,wndStyle,FALSE);
	
	if (wndRect.left<0)
	{
		wndRect.right -= wndRect.left;
		wndRect.left = 0;
	}
	
	if (wndRect.top<0)
	{
		wndRect.bottom -= wndRect.top;
		wndRect.top = 0;
	}

	m_wndSize.cx = wndRect.Width();
	m_wndSize.cy = wndRect.Height();
	
	// Create Background Brush
	m_bkBrush.CreateSolidBrush(BACKGROUND_COLOR);

	// Register Class                   
	HICON hIcon = AfxGetApp()->LoadIcon(IDI_GRAPHOUTWND);
	HBRUSH hBrush = (HBRUSH)m_bkBrush.m_hObject;
	HCURSOR hCursor = AfxGetApp()->LoadStandardCursor(IDC_ARROW);
	CString className = AfxRegisterWndClass(CS_NOCLOSE,
											hCursor,
											hBrush,
											hIcon);

	// Create MDIChild Window
	if (!CMDIChildWnd::Create(className,		// Registered Class Name
							  (LPCSTR)wndTitle,	// Window Title
							  wndStyle,			// Window Style
							  wndRect,			// Window Rectangle
                              pParentFrame)) 	// pParentWnd = main frame window	
		return FALSE;											 
		
	return TRUE;
}			 	   

void CGraphCoreWnd::OnDestroy()
{
	// Destroy GDI Objects
	m_bkBrush.DeleteObject();
	
	CMDIChildWnd::OnDestroy();
}


/////////////////////////////////////////////////////////////////////////////                       
// Message Handlers 

BEGIN_MESSAGE_MAP(CGraphCoreWnd, CMDIChildWnd)
	//{{AFX_MSG_MAP(CGraphCoreWnd)
	ON_WM_PAINT()
	ON_WM_DESTROY()
	ON_WM_SIZE()
	ON_WM_INITMENU()
	ON_WM_SYSCOMMAND()
	ON_WM_WINDOWPOSCHANGING()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()



/////////////////////////////////////////////////////////////////////////////                       
// Public Operations
void CGraphCoreWnd::ClearAll()
{
	// Clear Core
	for (int i=0; i<m_coreSize; i++)
		m_coreStatus[i] = NOT_USED;

	// Erase Background              
	CRect clientArea;
	CClientDC dc(this);

	GetClientRect(clientArea);
	dc.FillRect(clientArea,&m_bkBrush);
}
              
void CGraphCoreWnd::AccessedBy(CORE_ADDRESS addr, UINT nProg)
{
	if ((addr<(CORE_ADDRESS)m_coreSize)
		&& ((nProg==1) || (nProg==2)))
	{
		BYTE status = (nProg==1) ? ACCESSED_P1 : ACCESSED_P2;
		COLORREF color = (nProg==1) ? ACCESSED_P1_COLOR : ACCESSED_P2_COLOR;
		
		// Update Core
		m_coreStatus[addr] = status;
		
		// Paint Core Cell
		PaintCell(addr,color);
	}
}
                                              
void CGraphCoreWnd::ExecutedBy(CORE_ADDRESS addr, UINT nProg)
{
	if ((addr<(CORE_ADDRESS)m_coreSize)
		&& ((nProg==1) || (nProg==2)))
	{
		BYTE status = (nProg==1) ? EXECUTED_P1 : EXECUTED_P2;
		COLORREF color = (nProg==1) ? EXECUTED_P1_COLOR : EXECUTED_P2_COLOR;
		
		// Update Core
		m_coreStatus[addr] = status;
		
		// Paint Core Cell
		PaintCell(addr,color);
	}
}
                                

/////////////////////////////////////////////////////////////////////////////                                
// Painting
void CGraphCoreWnd::OnPaint()
{                   
	CRect paintRect;
	GetUpdateRect(paintRect);

	// Redraw Only Background
	EraseBackground(paintRect);
}
                    
void CGraphCoreWnd::EraseBackground(LPCRECT pRect)
{                   
	CRect paintRect;
	CClientDC dc(this);
	
	if (pRect==NULL)
		GetClientRect(paintRect);
	else
		paintRect = pRect;
	
	// Redraw Only Background
	dc.FillRect(paintRect,&m_bkBrush);
	ValidateRect(paintRect);
}

void CGraphCoreWnd::PaintCell(CORE_ADDRESS addr, COLORREF color)
{
	CClientDC dc(this);
	
	// Calculate Cell Position
	int yCell = addr / m_horzCells;
	int xCell = addr % m_horzCells;
	
	// Do Not Paint Whole Cell;
	// Leave the Top-Left Border Background Colored
	CRect cellRect;
	cellRect.top = yCell*CELL_HEIGHT + 1;
	cellRect.left = xCell*CELL_WIDTH + 1;
	cellRect.bottom = cellRect.top + CELL_HEIGHT - 1;
	cellRect.right = cellRect.left + CELL_WIDTH - 1;
	
	// Create Background Brush
	CBrush inkBrush;
	inkBrush.CreateSolidBrush(color);

	// Fill Selected Area
	dc.FillRect(cellRect,&inkBrush);
	
	// Free GDI Memory
	inkBrush.DeleteObject();	
}


/////////////////////////////////////////////////////////////////////////////
// Disabling Resizing Operations
void CGraphCoreWnd::OnSize(UINT nType, int cx, int cy)
{
	// Do Not Resize!!
	if ((IsIconic())
		|| (nType==SIZE_MINIMIZED))
		CMDIChildWnd::OnSize(nType, cx, cy);
}

// System Resizing
void CGraphCoreWnd::OnInitMenu(CMenu* pMenu)
{
	CMDIChildWnd::OnInitMenu(pMenu);
	
    // Disable Resizing Operations
    CMenu* pSysMenu = GetSystemMenu(FALSE);
    pSysMenu->RemoveMenu(SC_SIZE,MF_BYCOMMAND);
    pSysMenu->RemoveMenu(SC_MAXIMIZE,MF_BYCOMMAND);
}

void CGraphCoreWnd::OnSysCommand(UINT nID, LPARAM lParam)
{
	nID &= 0xfff0;
	
	// Do Not Resize!!
	if ((nID==SC_SIZE) 
		|| (nID==SC_MAXIMIZE))
		return;
		
	CMDIChildWnd::OnSysCommand(nID, lParam);
}

void CGraphCoreWnd::OnWindowPosChanging(WINDOWPOS FAR* lpwndpos)
{
	if ((!IsIconic())						// No Minimizing
		&& (lpwndpos->cx!=0)
		&& (lpwndpos->cy!=0)				// Resizing
		&& (lpwndpos->cx!=36)
		&& (lpwndpos->cy!=36)				// Icon Size
		&& (lpwndpos->cx!=m_wndSize.cx)
		&& (lpwndpos->cy!=m_wndSize.cy))	// Modifying Size
	{	
		lpwndpos->cx = m_wndSize.cx;
		lpwndpos->cy = m_wndSize.cy;
		lpwndpos->flags |= SWP_NOMOVE;
	}
	
	CMDIChildWnd::OnWindowPosChanging(lpwndpos);
}

