// cwmars.h : M.A.R.S. interface
/////////////////////////////////////////////////////////////////////////////


#include "time.h"
#include "cwmars.h"


IMPLEMENT_DYNCREATE(CRedcodeSimulator,CObject)


/////////////////////////////////////////////////////////////////////////////
// M.A.R.S. Construction/Destruction
CRedcodeSimulator::CRedcodeSimulator()
{
	m_coreSize = 0;
	m_programsGap = 0;
	m_battleSteps = 0;
	m_maxTasks = 0;

	// Try To Allocate The Core
	DWORD coreSize = MAX_CORE_SIZE;
	coreSize *= sizeof(CCoreCell);
	m_hCore = ::GlobalAlloc(GPTR,coreSize);
	
	if (m_hCore!=NULL)
		m_pCore = (CCoreCell __far*)::GlobalLock(m_hCore);
	else
		m_pCore = NULL;
	
	// Erase Data
	ResetScores();
	srand((unsigned)time(NULL));

	m_battles = 0;
	m_stepsDone = 0;
	m_battlesDone = 0;
	m_stepsPerRound	= 100;	// 1 Round = 100 Steps

	// Invalidate Pointers
	m_pTextWnd = NULL;
	m_pGraphWnd = NULL;
}

BOOL CRedcodeSimulator::InitializeMARS(int nCoreSize,
									   int nWarBattles,
									   int nBattleSteps,
									   int nProgramsGap,
									   int nProgramMaxTasks)
{
	m_coreSize = nCoreSize;
	m_warBattles = nWarBattles;
	m_battleSteps = nBattleSteps;
	m_programsGap = nProgramsGap;
	m_maxTasks = nProgramMaxTasks;

	return (m_pCore!=NULL);
		
}

CRedcodeSimulator::~CRedcodeSimulator()
{
	if (m_hCore!=NULL)
	{
		::GlobalUnlock(m_hCore);
		::GlobalFree(m_hCore);
	}
}


/////////////////////////////////////////////////////////////////////////////
// M.A.R.S. Pre-Battle Operations
void CRedcodeSimulator::ResetScores()
{
	m_winsP1 = 0;
	m_winsP2 = 0;
	m_lossesP1 = 0;
	m_lossesP2 = 0;
	m_drawsP1 = 0;
	m_drawsP2 = 0;
	m_scoreP1 = 0;
	m_scoreP2 = 0;
}

BOOL CRedcodeSimulator::LoadFirstProgram(const char* pPathName)
{
	m_pathP1 = pPathName;
	m_codeP1.ClearProgram();
	return LoadCodedProgram(pPathName,1);
}

BOOL CRedcodeSimulator::LoadSecondProgram(const char* pPathName)
{
	m_pathP2 = pPathName;
	m_codeP2.ClearProgram();
	return LoadCodedProgram(pPathName,2);
}

BOOL CRedcodeSimulator::LoadCodedProgram(const char* pPathName, UINT nProg)
{
	// Open File
	CFile file;
	if (!file.Open(pPathName,CFile::modeRead))
		return FALSE;
		
	// Create Related Archive
	CArchive archive(&file,CArchive::load);
	
	// Serialize CCoreProgram Object
	switch (nProg)
	{
		case 1	:	m_codeP1.Serialize(archive);
					return TRUE;

		case 2	:	m_codeP2.Serialize(archive);
					return TRUE;
		
		default	:	return FALSE;
	}                            
	
	// Should Never Be Executed...
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
// M.A.R.S. Fighting Operations
BOOL CRedcodeSimulator::LoadCodeIntoCore(UINT nProg)
{
	CDWordArray* pMyTasks = NULL;
	CDWordArray* pOtherTasks = NULL;
	CCoreProgram* pMyProgram = NULL;
	CORE_ADDRESS* pLoadAddress = NULL;
	
	switch (nProg)
	{
		case 1	:	pMyTasks = &m_tasksP1;
					pMyProgram = &m_codeP1;  
					pOtherTasks = &m_tasksP2;
					pLoadAddress = &m_loadAddressP1;
					break;

		case 2	:	pMyTasks = &m_tasksP2;
					pMyProgram = &m_codeP2;
					pOtherTasks = &m_tasksP1;
					pLoadAddress = &m_loadAddressP2;
					break;
	}

	if ((pMyTasks==NULL)
		|| (pMyProgram==NULL)
		|| (pOtherTasks==NULL))
	{		
		// Program Does Not Exist
		return FALSE;
	}
	
	// Select Load Address
	CORE_ADDRESS loadAddr;
	
	if (pOtherTasks->GetSize()==0)
	{
		// Random Address
		loadAddr = rand() % m_coreSize;
	}                                 
	else
	{
		// Minimal Gap Guaranteed
		loadAddr = rand() % m_coreSize;
		CORE_ADDRESS oppAddr = (CORE_ADDRESS)pOtherTasks->GetAt(0);
	    CORE_ADDRESS programsGap = (CORE_ADDRESS)abs(oppAddr - loadAddr);
		
		if (programsGap<(CORE_ADDRESS)m_programsGap)
		{
			if (loadAddr<oppAddr)                                   
				loadAddr = (loadAddr - m_programsGap) % m_coreSize;
			else
				loadAddr = (loadAddr + m_programsGap) % m_coreSize;
		}
	}

	// Copy Program Code
	for (int i=0; i<pMyProgram->GetProgramLength(); i++)
		m_pCore[(loadAddr + i) % m_coreSize] = *(pMyProgram->GetStatement(i));
                         
	// Done
	*pLoadAddress = loadAddr;
	pMyTasks->Add((CORE_ADDRESS)
				((loadAddr + pMyProgram->GetStartingPoint()) % m_coreSize));	
	
	return TRUE;
}

BOOL CRedcodeSimulator::PrepareBattleField()
{
	// Reset Core
	CCoreCell emptyCell;
	for (int i=0; i<m_coreSize; i++)
		m_pCore[i] = emptyCell;
	
	// Reset Old Task Lists
	m_tasksP1.RemoveAll();
	m_tasksP2.RemoveAll();
	
	// Load Programs Into The Core
	if ((!LoadCodeIntoCore(1))
		|| (!LoadCodeIntoCore(2)))
	{	
		// Failed
		m_isBattleAborted = TRUE;

		if (m_pTextWnd!=NULL)
		{
			m_pTextWnd->InsertLine("Battle Aborted:");
			m_pTextWnd->InsertLine(" * Can't Load Program(s) into Core.");
		}
        
		return FALSE;
	}

	// Select Starting Program
	m_firstPlayerID = m_currProgramID = 1 + rand()%2;
	
	// Init Battle Flags
	m_stepsDone = 0;
	m_isBattleAborted = FALSE;
	m_isProgramKilled = FALSE;
	
	// Reset Graphic Output
	if (m_pGraphWnd!=NULL)
		m_pGraphWnd->ClearAll();
		
	return TRUE;
}

BOOL CRedcodeSimulator::ResumeCurrentBattle()
{
	UINT currRoundSteps = 0;
	CDWordArray* pTasks = (m_currProgramID==1) ? &m_tasksP1 : &m_tasksP2;	
	
	while ((!m_isProgramKilled)
		&& (m_stepsDone<(UINT)m_battleSteps)
		&& (currRoundSteps<m_stepsPerRound))
	{
		ExecuteNextStatement(m_currProgramID);
		m_isProgramKilled = (pTasks->GetSize()==0);						 
		
		if (!m_isProgramKilled)
		{
			// Switch Programs & Task Lists
			m_currProgramID = 3 - m_currProgramID;
			pTasks = (pTasks==&m_tasksP1) ? &m_tasksP2 : &m_tasksP1;
			
			ExecuteNextStatement(m_currProgramID);
			m_isProgramKilled = (pTasks->GetSize()==0);
		}
		
		m_stepsDone++;
		currRoundSteps++;
	}

	if (m_isProgramKilled)
	{
		if (m_currProgramID==1)
		{
			// Program 1 Died, Program 2 Won
			m_winsP2++;
			m_lossesP1++;
			m_scoreP1 += CORE_WAR_LOSS_SCORE;
			m_scoreP2 += CORE_WAR_WIN_SCORE;
			WriteBattleResults(1);		
		}

		if (m_currProgramID==2)
		{
			// Program 2 Died, Program 1 Won
			m_winsP1++;
			m_lossesP2++;
			m_scoreP1 += CORE_WAR_WIN_SCORE;
			m_scoreP2 += CORE_WAR_LOSS_SCORE;
			WriteBattleResults(2);		
		}

		m_battlesDone++;

		// Another Battle?
		if (m_battlesDone<m_battles)
			PrepareBattleField(); 
		else
			WriteFinalReport();

		return (m_battlesDone<m_battles);	// FALSE only if last battle
	}            
	else
	{
		if (m_stepsDone>=(UINT)m_battleSteps)
		{
			// Draw!
			m_drawsP1 += 1;
			m_drawsP2 += 1;
			m_scoreP1 += CORE_WAR_DRAW_SCORE;
			m_scoreP2 += CORE_WAR_DRAW_SCORE;
			WriteBattleResults(0);

			// Battle Ended...
			m_battlesDone++;
			
			// Another Battle?
			if (m_battlesDone<m_battles)
				PrepareBattleField(); 
			else
				WriteFinalReport();
		
			return (m_battlesDone<m_battles);
		}
		else
		{
			// Round Ended
			return TRUE;
		}
	}
}

BOOL CRedcodeSimulator::StartBattle()
{
	// Reset Text Window
	if (m_pTextWnd!=NULL)
	{	
		m_pTextWnd->ClearAll();
	
		// Write Program Names
		char buffer[256];
		char drive[_MAX_DRIVE];
		char dir[_MAX_DIR];
		char fname[_MAX_FNAME];
		char ext[_MAX_EXT];
	
		// Extract 1st Player File Name From Path
		_splitpath(m_pathP1,drive,dir,fname,ext);

		strupr(fname);strupr(ext);
		sprintf(buffer,
				"P1 is ""%s%s"", %d instructions long",
				fname,ext,m_codeP1.GetProgramLength());		
		m_pTextWnd->InsertLine(buffer);
		
		// Extract 2nd Player File Name From Path
		_splitpath(m_pathP2,drive,dir,fname,ext);

		strupr(fname);strupr(ext);
		sprintf(buffer,
				"P2 is ""%s%s"", %d instructions long",
				fname,ext,m_codeP2.GetProgramLength());		
		m_pTextWnd->InsertLine(buffer);
	
		// Insert an Empty Line
		m_pTextWnd->InsertLine("");
	}              
	
	ResetScores();

	// Single Battle Option
	m_battles = 1;
	m_battlesDone = 0;

	return PrepareBattleField();
}

BOOL CRedcodeSimulator::StartCoreWar()
{
	// Reset Text Window
	if (m_pTextWnd!=NULL)
	{	
		m_pTextWnd->ClearAll();
	
		// Write Program Names
		char buffer[256];
		char drive[_MAX_DRIVE];
		char dir[_MAX_DIR];
		char fname[_MAX_FNAME];
		char ext[_MAX_EXT];
	
		// Extract 1st Player File Name From Path
		_splitpath(m_pathP1,drive,dir,fname,ext);

		strupr(fname);strupr(ext);
		sprintf(buffer,
				"P1 is ""%s%s"", %d instructions long",
				fname,ext,m_codeP1.GetProgramLength());		
		m_pTextWnd->InsertLine(buffer);
		
		// Extract 2nd Player File Name From Path
		_splitpath(m_pathP2,drive,dir,fname,ext);

		strupr(fname);strupr(ext);
		sprintf(buffer,
				"P2 is ""%s%s"", %d instructions long",
				fname,ext,m_codeP2.GetProgramLength());		
		m_pTextWnd->InsertLine(buffer);
	
		// Insert an Empty Line
		m_pTextWnd->InsertLine("");
	}              
	
	ResetScores();

	// Core War Battles
	m_battles = m_warBattles;
	m_battlesDone = 0;

	return PrepareBattleField();
}

BOOL CRedcodeSimulator::ExecuteNextStatement(UINT nProg)
{
	CDWordArray* pTasks = NULL;
	
	switch (nProg)
	{
		case 1	:	pTasks = &m_tasksP1;
					break;

		case 2	:	pTasks = &m_tasksP2;
					break;
	}

	if ((pTasks==NULL)
		|| (pTasks->GetSize()==0))
	{
		// No Task Found - Program Killed
		return FALSE;
	}
	
	// Get Current PC
	CORE_ADDRESS currPC = (CORE_ADDRESS)(pTasks->GetAt(0));
	pTasks->RemoveAt(0);
	
	// Retrieve Current Instruction
	CCoreCell __far* pCurrCell = &m_pCore[currPC];
	
	// Get Operands Locations
	CORE_ADDRESS opALocation = GetAOperandLocation(currPC);
	CORE_ADDRESS opBLocation = GetBOperandLocation(currPC);
	
	CCoreCell __far* pACell = &m_pCore[opALocation];
	CCoreCell __far* pBCell = &m_pCore[opBLocation];

	// Get Operands Values
	OP_DATA opAValue;
	
	// A operand is stored in m_opA_data only when
	// its addressing mode is  immediate, otherwise 
	// the default cell value is the B operand value.
	if (pCurrCell->GetOpAMode()==0) 
		opAValue = pACell->GetOpAValue();
	else
		opAValue = pACell->GetOpBValue();
		
	// B value is always stored in the B operand
	// value,  whatever  the addressing  mode is.
	OP_DATA opBValue = pBCell->GetOpBValue();
	
	// Increment PC
	currPC++;
	
	// Statement Execution
	BOOL stepDone = TRUE;
	switch (pCurrCell->GetCommand())	
	{
		case 0 :	// DAT B - Not Executable
					stepDone = FALSE;
					break;
		
		case 1 :	// MOV A,B               
					if (pCurrCell->GetOpAMode()==0)

						// Immediate Addressing Mode On A Operand
						pBCell->SetOpBValue(opAValue);
					else

						// Whole Cell Copied
						(*pBCell) = (*pACell);
					break;
		
		case 2 :	// ADD A,B	
					pBCell->SetOpBValue(opBValue + opAValue);		
					break;
					
		case 3 :	// SUB A,B
					pBCell->SetOpBValue(opBValue - opAValue);		
					break;

		case 4 :	// JMP A
					currPC = opALocation;
					break;

		case 5 :	// JMZ A,B	
					if (opBValue==0)
						currPC = opALocation;
					break;

		case 6 :	// JMN A,B
					if (opBValue!=0)
						currPC = opALocation;
					break;

		case 7 :	// DJN A,B
					pBCell->SetOpBValue(pBCell->GetOpBValue() - 1);
					
					if ((opBValue-1)!=0)
						currPC = opALocation;
						
					break;

		case 8 :	// CMP A,B                 
					BOOL equals;
					
					if ((pCurrCell->GetOpAMode()==0)
						|| (pCurrCell->GetOpBMode()==0))
						equals = (opAValue==opBValue);
					else
						equals = ((*pBCell)==(*pACell));
					
					if (equals)
						currPC++;
						
					break;
		
		case 9 :	// Instruction Doesn't Exist
					stepDone = FALSE;
					break;
		
		case 10 :	// SPL B
					// "+1" Represents Current Task
					if ((pTasks->GetSize()+1)<m_maxTasks)
					{
						// Create Another Task
						pTasks->Add((DWORD)opBLocation);
					}
					break;
		
		default :	// Instruction Doesn't Exist
					stepDone = FALSE;
					break;
	}
	
	// Store Current Tasks
	if (stepDone)
		pTasks->Add((DWORD)(currPC % m_coreSize));
	
	// Update Graphic Output Window
	if (m_pGraphWnd!=NULL)
	{
		m_pGraphWnd->ExecutedBy(currPC-1,nProg);
		
		// Operand A Location Access
		if ((stepDone)
			&& (opALocation!=(currPC-1))
			&& (pCurrCell->GetCommand()!=10))	// SPL doesn't use operand A
			m_pGraphWnd->AccessedBy(opALocation,nProg);

		// Operand B Location Access
		if ((stepDone)
			&& (opBLocation!=(currPC-1))
			&& (pCurrCell->GetCommand()!=4))	// JMP doesn't use operand B
			m_pGraphWnd->AccessedBy(opBLocation,nProg);
	}
	
	return stepDone;
}

CORE_ADDRESS CRedcodeSimulator::GetAOperandLocation(CORE_ADDRESS currPC)
{
	CORE_ADDRESS absLocation;
	CCoreCell __far* pCell = &m_pCore[currPC];
	
	switch (pCell->GetOpAMode())
	{
		case 0	:	// Immediate
					absLocation = currPC;
					break;
		
		case 1	:	// Rel. Direct
					absLocation = (currPC + pCell->GetOpAValue());
					break;
					
		case 2	:	// Rel. Indirect                  
					{
						CORE_ADDRESS opALoc = currPC + pCell->GetOpAValue();
						CCoreCell __far* pACell = &m_pCore[opALoc];

						absLocation = (opALoc + pACell->GetOpBValue());
		            }
		            break;
		            
		case 3	:	// Autodecremented	
					{
						CORE_ADDRESS opALoc = currPC + pCell->GetOpAValue();
						CCoreCell __far* pACell = &m_pCore[opALoc];
						
						// Execute Pre-Decrement
						pACell->SetOpBValue(pACell->GetOpBValue() - 1);
						absLocation = (opALoc + pACell->GetOpBValue());
					}                
					break;
	}

	return (absLocation % m_coreSize);
}	

CORE_ADDRESS CRedcodeSimulator::GetBOperandLocation(CORE_ADDRESS currPC)
{
	CORE_ADDRESS absLocation;
	CCoreCell __far* pCell = &m_pCore[currPC];
	
	switch (pCell->GetOpBMode())
	{
		case 0	:	// Immediate
					absLocation = currPC;
					break;
		
		case 1	:	// Rel. Direct
					absLocation = (currPC + pCell->GetOpBValue());
					break;
					
		case 2	:	// Rel. Indirect
					{
						CORE_ADDRESS opBLoc = currPC + pCell->GetOpBValue();
						CCoreCell __far* pBCell = &m_pCore[opBLoc];

						absLocation = (opBLoc + pBCell->GetOpBValue());
	        	    }                
	        	    break;
	        	    
		case 3	:	// Autodecremented	              
					{
						CORE_ADDRESS opBLoc = currPC + pCell->GetOpBValue();
						CCoreCell __far* pBCell = &m_pCore[opBLoc];
					
						// Execute Pre-Decrement
						pBCell->SetOpBValue(pBCell->GetOpBValue() - 1);
						absLocation = (opBLoc + pBCell->GetOpBValue());
					}
					break;
	}

	return (absLocation % m_coreSize);
}	


/////////////////////////////////////////////////////////////////////////////
// M.A.R.S. Post-Battle Operations
int CRedcodeSimulator::GetPlayerOneScore(int* pWins,
										 int* pDraws,
										 int* pLosses)	
{
	// Player One Scoring
	if (pWins!=NULL)
		*pWins = m_winsP1;
	
	if (pDraws!=NULL)
		*pDraws = m_drawsP1;

	if (pLosses!=NULL)
		*pLosses = m_lossesP1;

	return m_scoreP1;
}

int CRedcodeSimulator::GetPlayerTwoScore(int* pWins,
										 int* pDraws,
										 int* pLosses)	
{
	// Player One Scoring
	if (pWins!=NULL)
		*pWins = m_winsP2;
	
	if (pDraws!=NULL)
		*pDraws = m_drawsP2;

	if (pLosses!=NULL)
		*pLosses = m_lossesP2;

	return m_scoreP2;
}

/////////////////////////////////////////////////////////////////////////////
// Text Output
void CRedcodeSimulator::WriteBattleResults(UINT nWinnerProg)
{
	if (m_pTextWnd==NULL)
		return;
	
	char buffer[256];
	sprintf(buffer,"Battle no. %d of %d:",m_battlesDone+1,m_battles);
	m_pTextWnd->InsertLine(buffer);
	
	sprintf(buffer,"P1 was loaded at %u",m_loadAddressP1);
	m_pTextWnd->InsertLine(buffer);

	sprintf(buffer,"P2 was loaded at %u",m_loadAddressP2);
	m_pTextWnd->InsertLine(buffer);

	sprintf(buffer,"First move was to P%d",m_firstPlayerID);
	m_pTextWnd->InsertLine(buffer);                       
	
	if ((nWinnerProg==1)
		|| (nWinnerProg==2))
	{
		sprintf(buffer,
				"Battle Response: P%d won after %u steps.",
				nWinnerProg,m_stepsDone);
		m_pTextWnd->InsertLine(buffer);
	}
	else
	{
		// No Winner Recognized
		m_pTextWnd->InsertLine("Battle Response: draw.");
    }

	m_pTextWnd->InsertLine("");
}

void CRedcodeSimulator::WriteFinalReport()
{
	if (m_pTextWnd==NULL)
		return;
	
	char buffer[256];
	m_pTextWnd->InsertLine("*** Final Report ***");
	sprintf(buffer,"P1 scored %d points",m_scoreP1);
	m_pTextWnd->InsertLine(buffer);

	sprintf(buffer,
			"wins %d, draws %d, losses %d",
			m_winsP1,m_drawsP1,m_lossesP1);
	m_pTextWnd->InsertLine(buffer);

	sprintf(buffer,"P2 scored %d points",m_scoreP2);
	m_pTextWnd->InsertLine(buffer);

	sprintf(buffer,
			"wins %d, draws %d, losses %d",
			m_winsP2,m_drawsP2,m_lossesP2);
	m_pTextWnd->InsertLine(buffer);

	m_pTextWnd->InsertLine("");          
	
	if (m_scoreP1>m_scoreP2)
		m_pTextWnd->InsertLine("The winner is P1");

	if (m_scoreP1<m_scoreP2)
		m_pTextWnd->InsertLine("The winner is P2");

	if (m_scoreP1==m_scoreP2)
		m_pTextWnd->InsertLine("No winner!");
}

void CRedcodeSimulator::BattleAborted()
{
	m_isBattleAborted = TRUE;
	
	if (m_pTextWnd==NULL)
		return;

	m_pTextWnd->InsertLine("Battle Stopped:");
	m_pTextWnd->InsertLine(" * User Abort Message Received.");
	m_pTextWnd->InsertLine("");

}